# Copyright (c) 2010 Outpost Embedded, LLC
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Outpost Embedded trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Outpost Embedded, LLC.

import os, sys, tempfile, Queue, time
from subprocess import Popen, PIPE, STDOUT
from socket import socket, AF_UNIX, SOCK_STREAM

from pyudev import Context, Monitor
from pyudev.pyqt4 import QUDevMonitorObserver

from rapidwriter.backends.base import BaseDisk, BaseBackend
from rapidwriter.backends.linux import diskdaemon
from rapidwriter.logger import (
  log_debug,
  log_info,
  log_warning,
  log_error,
  log_traceback,
)
from rapidwriter.exceptions import (
  DeviceError,
  DeviceBusyError,
  DeviceNotFoundError,
  SubprocessError,
  PrivilegeEscalationError,
)
from rapidwriter.util import CallbackList


# FIXME: Do we need to worry about locking device objects returned from
# pyudev?  Are they thread safe?


_escalation_prog = 0

def get_escalation_args(args, msg):
    global _escalation_prog

    if _escalation_prog != 0:
        prog = _escalation_prog
    else:
        if os.getuid() == 0:
            prog = None
        else:
            have_gksu = (os.system('which gksu') == 0)
            have_kdesudo = (os.system('which kdesudo') == 0)
            have_pkexec = (os.system('which pkexec') == 0)

            if (
              'DISPLAY' not in os.environ
              or 'RAPIDWRITER_FORCE_SUDO' in os.environ
            ):
                prog = 'sudo'

            elif (
              'DESKTOP_SESSION' in os.environ
              and os.environ['DESKTOP_SESSION'] == 'gnome'
              and have_gksu
            ):
                # Running under GNOME and gksu is available.
                prog = 'gksu'

            elif (
              'KDE_FULL_SESSION' in os.environ
              and have_kdesudo
            ):
                # Running under KDE and kdesudo is available.
                prog = 'kdesudo'

            else:
                # Unknown desktop environment, just pick anything
                # that is available.
                if have_gksu:
                    prog = 'gksu'
                elif have_kdesudo:
                    prog = 'kdesudo'
                elif have_pkexec:
                    prog = 'pkexec'
                else:
                    raise PrivilegeEscalationError()

        # Cache result of detection so we don't have to redetect on the next
        # call.
        _escalation_prog = prog

    if prog is None:
        final_args = list(args)

    elif prog == 'sudo':
        final_args = ['sudo', '--']
        final_args.extend(args)

    elif prog == 'gksu':
        final_args = ['gksu', '--sudo-mode', '--message', msg, '--']
        final_args.extend(args)

    elif prog == 'kdesudo':
        final_args = ['kdesudo', '-d', '--comment', msg, '--']

        # kdesudo does some really braindead argument parsing (it strips quotes
        # from arguments), so we have to work around that here.  We only try to
        # fix things up for shell commands, which we know we can add spaces to
        # fool kdesudo's parser without invalidating the command.  This is
        # brittle, but I'm not sure what else to do about it at the moment.
        if args[0:2] == ['sh', '-c']:
            final_args.append(args[0])
            final_args.append(args[1])
            arg = args[2]
            if (
              (arg.startswith('"') and arg.endswith('"'))
              or (arg.startswith("'") and arg.endswith("'"))
            ):
                arg = ' %s ' % arg
            final_args.append(arg)
            final_args.extend(args[3:])
        else:
            final_args.extend(args)

    elif prog == 'pkexec':
        final_args = ['pkexec']
        final_args.extend(args)

    if prog is None:
        log_info('no escalation required, running as root')
    else:
        log_info('using %s for privilege escalation', prog)

    log_debug('final_args: %r', final_args)

    return final_args


class UDev(object):
    _subsystem = None
    _context = None
    _monitor = None
    _observer = None

    def __init__(self, subsystem):
        self._subsystem = subsystem

        self.device_added_callbacks = CallbackList()
        self.device_removed_callbacks = CallbackList()
        self.device_changed_callbacks = CallbackList()
        self.device_moved_callbacks = CallbackList()

        self._context = Context()
        self._monitor = Monitor.from_netlink(self._context)
        self._monitor.filter_by(subsystem = subsystem)
        self._observer = QUDevMonitorObserver(self._monitor)

        self._observer.deviceAdded.connect(self.device_added_callbacks.call)
        self._observer.deviceRemoved.connect(self.device_removed_callbacks.call)
        self._observer.deviceChanged.connect(self.device_changed_callbacks.call)
        self._observer.deviceMoved.connect(self.device_moved_callbacks.call)

    def start(self):
        self._monitor.start()

    def iter_devices(self):
        for device in self._context.list_devices().match_subsystem(self._subsystem):
            yield device


class DiskDaemonClient(object):
    _workdir = None
    _socket = None

    def start(self):
        self._workdir = tempfile.mkdtemp()
        socket_path = os.path.join(self._workdir, diskdaemon.SOCKET_NAME)

        final_args = get_escalation_args(
          args = [
            sys.executable,
            os.path.abspath(diskdaemon.__file__),
            self._workdir,
            str(os.getuid()),
          ],
          msg = (
            'Administrative access is required for '
            'disk management functions.'
          ),
        )

        # FIXME: If something goes wrong and we never connect and disconnect
        # to the daemon process, it will continue to run in the background.
        # We could fix this perhaps by always trying to connect, but a better
        # way to fix it would be for the daemon to automatically exit if no
        # connection is received within a reasonable amount of time.

        popen = Popen(final_args, stdout = PIPE, stderr = STDOUT)
        try:
            popen.wait()
            output = popen.stdout.read()
        finally:
            popen.stdout.close()
        if popen.returncode != 0:
            log_error(output)
            raise SubprocessError(popen.returncode, output)

        self._socket = socket(AF_UNIX, SOCK_STREAM)
        self._socket.connect(socket_path)

        self.setpath(sys.path)

    def stop(self):
        with open(os.path.join(self._workdir, diskdaemon.LOG_FILE), 'r') as f:
            log_content = f.read().strip()
        log_info('diskdaemon log:')
        log_info(log_content)
        self._socket.close()

    def _send_command(self, cmd, arg = None):
        if isinstance(cmd, unicode):
            cmd = cmd.encode(diskdaemon.PROTOCOL_ENCODING)

        if arg is not None:
            if not isinstance(arg, basestring):
                arg = unicode(arg)
            if isinstance(arg, unicode):
                arg = arg.encode(diskdaemon.PROTOCOL_ENCODING)
            full_command = '%s %s\n' % (cmd, arg)
        else:
            full_command = '%s\n' % cmd

        log_debug('diskdaemon: > %s', full_command)

        self._socket.send(full_command)
        chars = []
        while True:
            ch = self._socket.recv(1)
            if not ch:
                break
            if ch == '\n':
                break
            chars.append(ch)

        response = ''.join(chars)

        log_debug('diskdaemon: < %s', response)

        parts = response.split(' ', 1)
        if len(parts) == 1:
            response_type = parts[0]
            arg = None
        else:
            response_type, arg = parts

        response_type = response_type.decode(diskdaemon.PROTOCOL_ENCODING)
        if response_type not in (u'OK', u'ERROR'):
            log_error('bad response type: %r', response_type)

        if response_type == u'ERROR':
            if arg is None:
                log_error('bad error response (no arg)')
            else:
                parts = arg.split(' ', 1)
                if len(parts) == 1:
                    error_type = parts[0]
                    error_msg = None
                else:
                    error_type = parts[0]
                    error_msg = parts[1].decode(diskdaemon.PROTOCOL_ENCODING)
                error_type = error_type.decode(diskdaemon.PROTOCOL_ENCODING)

                if error_type == u'INVALID-COMMAND':
                    raise AssertionError(error_msg)
                elif error_type == u'DEVICE-BUSY':
                    raise DeviceBusyError(error_msg)
                elif error_type == u'DEVICE-NOT-FOUND':
                    raise DeviceNotFoundError(error_msg)
                else:
                    # FIXME: Is DeviceError right here?
                    raise DeviceError(error_msg)

        else:
            return arg

    ###

    def getpid(self):
        return int(self._send_command('GETPID'))

    def setpath(self, arg):
        self._send_command('SETPATH', repr(arg))

    def partition(self, arg):
        self._send_command('PARTITION', arg)

    def mkfs(self, arg):
        self._send_command('MKFS', arg)

    def bootloader(self, arg):
        self._send_command('BOOTLOADER', arg)

    def mount(self, arg):
        return self._send_command('MOUNT', arg).decode(
          diskdaemon.PROTOCOL_ENCODING,
        )

    def umount(self, arg):
        self._send_command('UMOUNT', arg)


class LinuxDisk(BaseDisk):
    udev = None
    device = None
    client = None

    fd = None

    def __init__(self, udev, device, client):
        super(LinuxDisk, self).__init__()
        self.udev = udev
        self.device = device
        self.client = client

    def __repr__(self):
        return '%s(%r, %r, %r)' % (
          type(self).__name__,
          self.udev,
          self.device,
          self.client,
        )

    ###

    def get_name(self):
        return os.path.basename(self.get_device_path())

    def get_description(self):
        return u'%s %s' % (self.get_manufacturer(), self.get_model())

    def get_size(self):
        # FIXME: Is using 512 here always correct?  What about disks with a
        # different physical block size?
        return (long(self.device.attributes['size']) * 512)

    def get_device_path(self):
        return self.device.device_node

    def is_removable(self):
        return (self.device.attributes['removable'] == '1')

    ###

    def start(self):
        # Keep an open file descriptor to the device to avoid a situation where
        # it is removed and another device is assigned the same name without us
        # noticing (unlikely, but distastrous).  An open file descriptor
        # prevents the kernel from reusing the device name.
        self.fd = os.open(self.get_device_path(), os.O_RDONLY)

    def stop(self):
        os.close(self.fd)

    ###

    def get_manufacturer(self):
        s = self.device.parent.attributes['vendor']
        return s.decode('utf-8', 'replace').strip()

    def get_model(self):
        s = self.device.parent.attributes['model']
        return s.decode('utf-8', 'replace').strip()

    def get_partition_device(self):
        for child in self.device.children:
            if (
              child.subsystem == 'block'
              and child.attributes['partition'] == '1'
            ):
                return child

        raise DeviceNotFoundError()

    def get_partition_path(self):
        return self.get_partition_device().device_node

    ###

    def create_partition_table(self):
        self.umount_all()
        self.client.partition(self.get_device_path())
        self.settle_partitions()

    def settle_partitions(self):
        queue = Queue.Queue()

        self.udev.device_added_callbacks.add(queue.put)
        try:
            self.udev.device_removed_callbacks.add(queue.put)
            try:
                settled = False
                count = 0

                log_info('waiting for device to settle')

                for index in range(10):
                    time.sleep(1)

                    removed = False

                    try:
                        device = self.get_partition_device()
                    except DeviceNotFoundError:
                        log_debug('device is not found')
                        device = None

                    if device is None:
                        # Reset the count.
                        count = 0

                        # Clear the queue.
                        while True:
                            try:
                                queue.get(block = False)
                            except Queue.Empty:
                                pass

                    else:
                        # Look to see if our device was removed in the last second.
                        while True:
                            try:
                                device_ = queue.get(block = False)
                            except Queue.Empty:
                                break
                            else:
                                log_debug('device was added or removed: %r', device_)
                                if device_ == device:
                                    removed = True

                                    # Clear the queue.
                                    while True:
                                        try:
                                            queue.get(block = False)
                                        except Queue.Empty:
                                            break

                                    break

                        if not removed:
                            count = count + 1
                            if count >= 2:
                                settled = True

                    log_info(
                      'after %u seconds, device %s settled',
                      (index + 1),
                      (settled and 'is' or 'is not'),
                    )

                    if settled:
                        break

            finally:
                self.udev.device_removed_callbacks.remove(queue.put)

        finally:
            self.udev.device_added_callbacks.remove(queue.put)

        if not settled:
            raise DeviceNotFoundError()

    def create_filesystem(self):
        self.client.mkfs(self.get_partition_path())

    def install_bootloader(self):
        self.client.bootloader(self.get_partition_path())

    def umount_all(self):
        mounted_devices = []
        with open('/proc/mounts', 'r') as f:
            for line in f:
                mounted_devices.append(line.split()[0])

        log_debug('mounted_devices = %r', mounted_devices)

        if self.device.device_node in mounted_devices:
            self.client.umount(self.device.device_node)

        for child in self.device.children:
            if child.subsystem == 'block':
                device_node = child.device_node
                if device_node in mounted_devices:
                    if device_node is None:
                        log_warning('device_node is None: %r', child)
                        continue
                    self.client.umount(child.device_node)

    def mount(self):
        mount_point = self.client.mount(self.get_partition_path())
        # Sadly, mount exits before the mount point is actually usable.  I
        # think this is a bug in the kernel.  Wait around for a few seconds
        # until the mount point is writable.
        writable = False
        for index in range(10):
            if os.access(mount_point, os.W_OK):
                writable = True
                break
            time.sleep(1)
        if not writable:
            raise DeviceError('mount point is not writable within 10 seconds')
        return mount_point

    def umount(self, mount_point):
        self.client.umount(self.get_partition_path())

    def sync(self):
        os.system('sync')


class LinuxBackend(BaseBackend):
    disks = []
    
    udev = None
    client = None

    _context = None
    _monitor = None
    _observer = None

    _workdir = None

    def __init__(self, *args, **kwargs):
        super(LinuxBackend, self).__init__(*args, **kwargs)

        self.disks = []

        self.udev = UDev('block')
        self.udev.device_added_callbacks.add(self._handle_device_added)
        self.udev.device_removed_callbacks.add(self._handle_device_removed)
        self.udev.device_changed_callbacks.add(self._handle_device_changed)
        self.udev.device_moved_callbacks.add(self._handle_device_moved)

        self.client = DiskDaemonClient()

    def start(self):
        self.client.start()
        self.udev.start()
        self._init_disks()

    def stop(self):
        # FIXME: Do we have to stop the UDev monitor?
        self.client.stop()

    def _init_disks(self):
        for device in self.udev.iter_devices():
            self._add_device(device)

    def _handle_device_added(self, device):
        log_info('udev: device added: %r', device)
        self._add_device(device)

    def _handle_device_removed(self, device):
        log_info('udev: device removed: %r', device)
        self._remove_device(device)

    def _handle_device_changed(self, device):
        # FIXME: Should we do anything when this happens?
        log_info('udev: device changed: %r', device)

    def _handle_device_moved(self, device):
        # FIXME: Should we do anything when this happens?
        log_info('udev: device moved: %r', device)

    def _add_device(self, device):
        # FIXME: scsi is appropriate, right?
        if (device.parent is not None) and (device.parent.subsystem == 'scsi'):
            log_info('udev: adding device: %r', device)
            try:
                disk = self._find_disk(device)
            except DeviceNotFoundError:
                disk = LinuxDisk(self.udev, device, self.client)
                self.disks.append(disk)
                self.add_disk(disk)

    def _remove_device(self, device):
        try:
            disk = self._find_disk(device)
        except DeviceNotFoundError:
            pass
        else:
            self.disks.remove(disk)
            self.remove_disk(disk)

    def _find_disk(self, device):
        for disk in self.disks:
            if disk.device == device:
                return disk
        raise DeviceNotFoundError()
